/*
 * Copyright (C) 2013 Matthew Dawson <matthew@mjdsystems.ca>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the distribution
 *     * Neither the name of the McMaster Centre for Software Certification nor the names
 *       of its contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
package ca.mcscert.jtet.cvc3generator;

import ca.mcscert.jtet.expression.VariableCollection;
import ca.mcscert.jtet.parsers.MatlabParser;
import ca.mcscert.jtet.expression.BooleanVariableType;
import ca.mcscert.jtet.expression.CVC3Generator;
import ca.mcscert.jtet.tablularexpression.Cell;
import ca.mcscert.jtet.tablularexpression.HierarchicalGridCheckerGenerator;

import java.util.ArrayList;
import java.util.ArrayDeque;
import java.util.Deque;
import java.util.List;

import org.apache.commons.lang3.StringUtils;

/**
 *
 * @author Matthew Dawson <matthew@mjdsystems.ca>
 */
final public class HierarchicalGridCVC3Generator implements HierarchicalGridCheckerGenerator {

    public HierarchicalGridCVC3Generator(VariableCollection variableDefinitions, int queryCountStart) {
        m_variableDefinitions = variableDefinitions;
        m_currentTCC = queryCountStart;
    }

    private String generateQueryPrefix() {
        if (!m_parentCellsCVC3Code.isEmpty()){
            String output = "(";
            output += StringUtils.join(m_parentCellsCVC3Code, " AND ");
            output += ") => ";
            return output;
        }
        return "";
    }

    private void outputCells() {
        // First is disjoint testing.
        m_output += "ECHO \"begin" + m_currentTCC + "\";\n";
        m_output += "PUSH;\n";

        String query = "QUERY ";
        query += generateQueryPrefix();
        query += "(" + StringUtils.join(m_currentDisjointQueries, " AND\n");
        query += ");";
        m_queries.add(query);
        m_output += query + "\n";

        m_output += "POP;\n";
        m_output += "ECHO \"end" + m_currentTCC + "\";\n";

        // Finished first TCC, increment counter.
        m_currentTCC++;

        // Next is complete.
        m_output += "ECHO \"begin" + m_currentTCC + "\";\n";
        m_output += "PUSH;\n";

        query = "QUERY ";
        query += generateQueryPrefix();
        query += "(" + StringUtils.join(m_currentCompleteQueries, " OR ");
        query += ");";
        m_queries.add(query);
        m_output += query + "\n";

        m_output += "POP;\n";
        m_output += "ECHO \"end" + m_currentTCC + "\";\n";

        // Finished other TCC, increment counter.
        m_currentTCC++;

        m_currentlyRunning = false;

        // And finally clear the running list of current queries.
        m_currentCompleteQueries.clear();
        m_currentDisjointQueries.clear();
    }

    @Override
    public void descendIntoGridFromCell(Cell cell) {
        if (m_currentlyRunning) {
            outputCells();
        }
        m_parentCellsCVC3Code.addFirst(MatlabParser.parseMatlabCode(m_variableDefinitions, cell.contents()).getCheckerOutput(m_CVC3Generator, BooleanVariableType.Type));
        m_currentlyRunning = true;
    }

    @Override
    public void ascendFromGrid() {
        if (m_currentlyRunning) {
            outputCells();
        }
        m_parentCellsCVC3Code.removeFirst();
    }

    @Override
    public void handleEdgeCell(Cell cell) {
        handleLeafCell(cell);
    }

    @Override
    public void handleLeafCell(Cell cell) {
        // First get the appropriate CVC3 code from matlab.
        String newCellCode = MatlabParser.parseMatlabCode(m_variableDefinitions, cell.contents()).getCheckerOutput(m_CVC3Generator, BooleanVariableType.Type);

        // Next form the disjoint queries and add to the list.
        for(String otherCellCode : m_currentCompleteQueries) {
            String newDisjointQuery = "NOT ( " + otherCellCode  + " AND " + newCellCode + " )";
            m_currentDisjointQueries.add(newDisjointQuery);
        }

        // Lastly add the cell to the list for the complete query
        m_currentCompleteQueries.add(newCellCode);
    }

    @Override
    public String getFinalString() {
        if (m_currentlyRunning) {
            outputCells();
        }
        return m_output;
    }

    public List<String> getFinalQueries() {
        if (m_currentlyRunning) {
            outputCells();
        }
        return m_queries;
    }

    Deque<String> m_parentCellsCVC3Code = new ArrayDeque<String>();

    List<String> m_currentDisjointQueries = new ArrayList<String>();
    List<String> m_currentCompleteQueries = new ArrayList<String>();
    
    boolean m_currentlyRunning = true;
    int m_currentTCC = 1;

    String m_output = "";
    List<String> m_queries = new ArrayList<String>();

    // To fix properly.
    VariableCollection m_variableDefinitions;
    CVC3Generator m_CVC3Generator = new CVC3Generator();
}
