/*
 * Copyright (C) 2013 Matthew Dawson <matthew@mjdsystems.ca>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the distribution
 *     * Neither the name of the McMaster Centre for Software Certification nor the names
 *       of its contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
package ca.mcscert.jtet.expression;

import java.util.HashSet;
import java.util.Map;
import java.util.Set;

/**
 *
 * @author Matthew Dawson <matthew@mjdsystems.ca>
 */
public class SMTLIBGenerator implements CheckerGenerator {
    @Override
    public String GenerateUnaryOperation(UnaryOperation op, String expression, VariableType usedType) {
        return "(" + SMTLIBGenerator.GetUnaryOperationSymbolFor(op) + " " + expression + ")";  //To change body of implemented methods use File | Settings | File Templates.
    }

    @Override
    public String GenerateBinaryOperation(BinaryOperation op, String lhsExp, String rhsExp, VariableType usedType) {
        return "(" + SMTLIBGenerator.GetBinaryOperationSymbolFor(op) + " " + lhsExp + " " + rhsExp + ")";
    }

    @Override
    public String GenerateLiteralValue(String value, VariableType requestedType) {
        if (requestedType instanceof RealVariableType) {
            if (value.contains(".")) {
                return value;
            } else {
                return value + ".0";
            }
        } else {
            throw new IllegalArgumentException("Can't cast literal to that type!");
        }
    }

    @Override
    public String GenerateVariablesDeclaration(VariableCollection variableCollection) {
        Map<String, Variable> vars = variableCollection.getVariables();
        Set<EnumerationVariableType> emittedEnumerations = new HashSet<EnumerationVariableType>();

        StringBuilder ret = new StringBuilder();
        // First generate all the enumeration types.
        for(Variable var : vars.values()) {
            if (var.type() instanceof EnumerationVariableType && !emittedEnumerations.contains(var.type())) {
                // For each enumeration, first emit the actual enumeration type name.
                EnumerationVariableType enumType = (EnumerationVariableType)var.type();
                ret.append("(declare-sort ")
                        .append(enumType.name())
                        .append(" 0)\n");

                // Next, emit the enumeration values, using the enumeration type.
                for(String enumValue : enumType.enumerationValues()) {
                    ret.append("(declare-fun ")
                            .append(enumValue)
                            .append(" () ")
                            .append(enumType.name())
                            .append(")\n");
                }

                // Restrict the enumeration values so that they are all considered different.
                ret.append("(assert (distinct");
                for(String enumValue : enumType.enumerationValues()) {
                    ret.append(" ")
                            .append(enumValue);
                }
                ret.append("))\n");

                // Finally, emit a function that ensure any enumeration variable can only be the values given.
                ret.append("(define-fun is")
                        .append(enumType.name())
                        .append(" ((val ")
                        .append(enumType.name())
                        .append(")) Bool (or");
                for(String enumValue : enumType.enumerationValues()) {
                    ret.append(" (= val ")
                            .append(enumValue)
                            .append(")");
                }
                ret.append("))\n");

                emittedEnumerations.add(enumType);
            }
        }

        // Then generate all the variables.
        for (Variable var : vars.values()) {
            ret.append("(declare-fun ")
               .append(var.outputName())
               .append(" () ")
               .append(GenerateVariableType(var.type()))
               .append(")\n");

            // If variable is an enumeration, make sure that it is limited to only enumeration values.
            if (var.type() instanceof EnumerationVariableType) {
                ret.append("(assert (is")
                        .append(((EnumerationVariableType) var.type()).name())
                        .append(" ")
                        .append(var.outputName())
                        .append("))\n");
            }
        }

        // Then generate the variable predicates.  This ensures any dependent variables are declared.
        for(Variable var : vars.values()) {
            if (var.type().subtypePredicate() != null) {
                ret.append("(assert ")
                        .append(var.type().subtypePredicate().getCheckerOutput(this, BooleanVariableType.Type))
                        .append(")\n");
            }
        }
        return ret.toString();
    }

    public static String GenerateVariableType(VariableType type) {
        if (type instanceof RealVariableType) {
            return "Real";
        } else if (type instanceof BooleanVariableType) {
            return "Bool";
        } else if (type instanceof EnumerationVariableType) {
            return ((EnumerationVariableType) type).name();
        } else {
            throw new IllegalArgumentException("SMT-LIB Doesn't handle this variable type (Class: " + type.getClass().getName() + ")");
        }
    }

    private static String GetBinaryOperationSymbolFor(BinaryOperation op) {
        switch (op) {
            case Addition:
                return "+";
            case Minus:
                return "-";
            case Multiplication:
                return "*";
            case Division:
                return "/";
            case Exponentiation:
                return "^";
            case GreaterThen:
                return ">";
            case GreaterThenEqual:
                return ">=";
            case LessThen:
                return "<";
            case LessThenEqual:
                return "<=";
            case Equals:
                return "=";
            case NotEquals:
                return "distinct";
            case BooleanAnd:
                return "and";
            case BooleanOr:
                return "or";
            case Implies:
                return "=>";
        }
        throw new RuntimeException("Should never happen!  Asked for unhandled op: " + op + "!");
    }

    private static String GetUnaryOperationSymbolFor(UnaryOperation op) {
        switch (op) {
            case Negation:
                return "not";
            case Minus:
                return "-";
        }
        throw new RuntimeException("Should never happen!  Asked for unhandled op: " + op + "!");
    }
}
