/*
 * Copyright (C) 2013 Matthew Dawson <matthew@mjdsystems.ca>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the distribution
 *     * Neither the name of the McMaster Centre for Software Certification nor the names
 *       of its contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
package ca.mcscert.jtet.expression;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

/**
 * VariableCollection stores a set of variables for use by expression parsers.
 * <p>
 * Expression parsers need an easy way to find all the valid variables for a given expression.  This class receives a
 * set of variables, which it then makes available.
 * <p>
 * It also provides an easy way to find all the enumeration constants available.  It mixes this together with the
 * variables, allowing the parser to quickly check if a given name is valid.
 *
 * @author Matthew Dawson
 */
final public class VariableCollection {
    /**
     * Constructs a VariableCollection with the given variables and enumerations.
     * <p>
     * The VariableCollection constructed here will have the variables given, along with all the enumeration constants
     * from the given set of enumerations.  Note that all enumeration constants must not conflict with each other or
     * any given variable.
     *
     * @param variables This variables to store in this collection.
     * @param enumerationTypes The set of enumerations for which the constants should be extracted.  May be null.
     */
    public VariableCollection(Map<String, Variable> variables, Set<EnumerationVariableType> enumerationTypes) {
        m_variables = variables;
        if (m_variables != null) {
            for(Map.Entry<String, Variable> var : m_variables.entrySet()) {
                m_variablesAndEnumeratedValues.put(var.getKey(), var.getValue());
            }
        }

        if (enumerationTypes != null) {
            for(EnumerationVariableType enumType : enumerationTypes) {
                for(String enumValue : enumType.enumerationValues()) {
                    if (m_variablesAndEnumeratedValues.containsKey(enumValue)) {
                        throw new IllegalArgumentException("Enumerated value conflicts with existing variable!");
                    }
                    m_variablesAndEnumeratedValues.put(enumValue, new Variable(enumValue, enumType));
                }
            }
        }
    }

    /**
     * Constructs a VariableCollection with the given variables and enumerations.
     * <p>
     * The VariableCollection constructed here will have the variables given.  No enumerations will be stored.
     *
     * @param variables This variables to store in this collection.
     */
    public VariableCollection(Map<String, Variable> variables) {
        this(variables, null);
    }

    /**
     * Returns a map containing all the known variables and enumeration constants.
     * <p>
     * Whatever is available at a given name, it is guaranteed to be a unique name.
     *
     * @return A map of variables and enumeration constants.
     */
    public Map<String, Variable> getVariablesAndEnumeratedValues() {
        return Collections.unmodifiableMap(m_variablesAndEnumeratedValues);
    }

    /**
     * Returns a map containing only the variables.
     * <p>
     * This ignores all the enumeration constants present.
     * @return A map of variables.
     */
    public Map<String, Variable> getVariables() {
        return Collections.unmodifiableMap(m_variables);
    }

    private final Map<String,Variable> m_variables;
    private final Map<String,Variable> m_variablesAndEnumeratedValues = new HashMap<String, Variable>();
}
