/*
 * Copyright (C) 2013 Matthew Dawson <matthew@mjdsystems.ca>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the distribution
 *     * Neither the name of the McMaster Centre for Software Certification nor the names
 *       of its contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
package ca.mcscert.jtet.smtlibchecker;

import ca.mcscert.jtet.expression.SMTLIBGenerator;
import ca.mcscert.jtet.expression.Variable;
import ca.mcscert.jtet.expression.VariableCollection;
import ca.mcscert.jtet.smtlibchecker.antlr.SMTLibOutParserBaseListener;
import ca.mcscert.jtet.smtlibchecker.antlr.SMTLibOutParserLexer;
import ca.mcscert.jtet.smtlibchecker.antlr.SMTLibOutParserListener;
import ca.mcscert.jtet.smtlibchecker.antlr.SMTLibOutParserParser;
import ca.mcscert.jtet.tablularexpression.HierarchicalGridCheckerWalkerGenerator;
import ca.mcscert.jtet.tablularexpression.HierarchicalGrid;
import org.antlr.v4.runtime.ANTLRInputStream;
import org.antlr.v4.runtime.CommonTokenStream;
import org.antlr.v4.runtime.misc.NotNull;
import org.antlr.v4.runtime.tree.ParseTree;
import org.antlr.v4.runtime.tree.ParseTreeWalker;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.Reader;
import java.io.UnsupportedEncodingException;
import java.io.Writer;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 *
 * @author Matthew Dawson <matthew@mjdsystems.ca>
 */
final public class CheckerRunner {
    public static List<CheckerRunnerResult> RunZ3(HierarchicalGrid grid, VariableCollection vars) {
        List<CheckerRunnerResult> ret = new ArrayList<CheckerRunnerResult>();

        // Get the queries
        HierarchicalGridSMTLIBGenerator queryGenerator = new HierarchicalGridSMTLIBGenerator(vars);
        HierarchicalGridCheckerWalkerGenerator.GenerateCheckerFromGrid(grid, queryGenerator);

        // Start Z3
        ProcessBuilder pb;
        Process p;
        try {
            pb = new ProcessBuilder("z3", "-smt2", "-in");
            p = pb.start();
        } catch(IOException ex) {
            throw new RuntimeException("Failed to start Z3!", ex);
        }

        // Input/Output refer to the direction from Z3's pov.  So output is Z3's stdout.
        BufferedReader z3Output;
        Writer z3Input;
        try {
            // Sigh ... Java has horrible naming conversions here ...
            z3Output = new BufferedReader(new InputStreamReader(p.getInputStream(), "ISO-8859-1"));
            z3Input = new OutputStreamWriter(p.getOutputStream());
        } catch(UnsupportedEncodingException ex) {
            throw new RuntimeException("Failed to get ISO-8859-1 charset for talking to Z3!", ex);
        }

        StringBuilder GetValueRequestBuilder = new StringBuilder();
        GetValueRequestBuilder.append("(get-value (");
        for(Variable var:vars.getVariables().values()) {
            GetValueRequestBuilder.append(var.name()).append(' ');
        }
        GetValueRequestBuilder.append("))\n");
        String GetValueRequestString = GetValueRequestBuilder.toString();

        try {
            String VarDef = new SMTLIBGenerator().GenerateVariablesDeclaration(vars);
            z3Input.write(VarDef);
            for(String query : queryGenerator.getFinalQueries()) {
                z3Input.write("(push 1)");
                z3Input.write(query);
                z3Input.write("(check-sat)\n");
                z3Input.flush();

                String res = z3Output.readLine();
                if(res.equals("sat")) { // Failed query.  Get a the result.
                    z3Input.write(GetValueRequestString);
                    z3Input.flush();

                    CheckerRunnerResult result = new CheckerRunnerResult();
                    result.Query = query;
                    result.SampleValues = parseSMTLIBOutput(z3Output);
                    ret.add(result);
                } else if(res.equals("unknown")) {
                    // Z3 failed to prove.  Throw a failure to check for now.
                    CheckerRunnerResult result = new CheckerRunnerResult();
                    result.Query = query;
                    result.SampleValues = null;
                    ret.add(result);
                } else if(!res.equals("unsat")) {
                    // Something went wrong
                    throw new IllegalStateException("Z3 outputed something wrong (" + res + ")!");
                }

                z3Input.write("(pop 1)\r");
            }
        } catch(IOException ex) {
            throw new RuntimeException("Something bad happened ...", ex);
        } finally {
            try {
                z3Input.write("(exit)\n");
                z3Input.flush();

                p.destroy();
                z3Input.close();
                z3Output.close();
            } catch(Exception e) {
                // Don't care here, so just print it out.
                System.out.println(e);
            }
        }

        return ret;
    }

    private static Map<String, String> parseSMTLIBOutput(Reader reader) throws IOException {
        if(reader.markSupported() == false) {
            throw new IllegalArgumentException("Reader must support marks!");
        }
        final Map<String, String> ret = new HashMap<String, String>();

        StringBuilder input = new StringBuilder();

        int bracketMismatch = 1; // This starts with a (, so we immediately have a a mismatch.
        char nextChar = getNextChar(reader);
        if (nextChar != '(') { // Should start with a ( for output.
            throw new IllegalStateException("Bad input, wanted a ( to start, got " + nextChar);
        }
        input.append(nextChar);
        while(bracketMismatch != 0) {
            nextChar = getNextChar(reader);
            switch (nextChar) {
                case '(':
                    bracketMismatch++;
                    break;
                case ')':
                    bracketMismatch--;
                    break;
            }
            input.append(nextChar);
        }

        // Finally, clear the new line, so the output is now clear.
        do {
            reader.mark(5);
            nextChar = getNextChar(reader);
        } while(isWhitespace(nextChar) && nextChar != '\n');
        if(!isWhitespace(nextChar)) {
            reader.reset();
        }

        ANTLRInputStream antlrInput = new ANTLRInputStream(input.toString());

        // create a lexer that feeds off of input CharStream
        SMTLibOutParserLexer lexer = new SMTLibOutParserLexer(antlrInput);

        // create a buffer of tokens pulled from the lexer
        CommonTokenStream tokens = new CommonTokenStream(lexer);

        // create a parser that feeds off the tokens buffer
        SMTLibOutParserParser parser = new SMTLibOutParserParser(tokens);

        ParseTree tree = parser.valuelist();
        ParseTreeWalker walker = new ParseTreeWalker();
        SMTLibOutParserListener listener = new SMTLibOutParserBaseListener() {
            @Override
            public void enterValue(@NotNull SMTLibOutParserParser.ValueContext ctx) {
                id = ctx.ID().getText();
                value = "";
            }

            @Override
            public void enterNegative(@NotNull SMTLibOutParserParser.NegativeContext ctx) {
                value = "(- ";
            }

            @Override
            public void exitSingle(@NotNull SMTLibOutParserParser.SingleContext ctx) {
                value += ctx.getText();
            }

            @Override
            public void exitFraction(@NotNull SMTLibOutParserParser.FractionContext ctx) {
                value += "(" + ctx.RATIONAL(0).getText() + " / " + ctx.RATIONAL(1).getText() + ")";
            }

            @Override
            public void exitNegative(@NotNull SMTLibOutParserParser.NegativeContext ctx) {
                value += ')';
            }

            @Override
            public void exitExpr(@NotNull SMTLibOutParserParser.ExprContext ctx) {
                ret.put(id, value);
            }

            String id = "";
            String value;
        };
        walker.walk(listener, tree);
        return ret;
    }

    private static char getNextChar(Reader reader) throws IOException {
        char[] c = new char[1];
        reader.read(c);
        return c[0];
    }

    private static boolean isWhitespace(char toTest) {
        if(toTest == ' ' ||
                toTest == '\n' ||
                toTest == '\r' ||
                toTest == '\t') {
            return true;
        }
        return false;
    }
}
