/*
 * Copyright (C) 2013 Matthew Dawson <matthew@mjdsystems.ca>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the distribution
 *     * Neither the name of the McMaster Centre for Software Certification nor the names
 *       of its contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
package ca.mcscert.jtet.expression.test;

import ca.mcscert.jtet.expression.EnumerationVariableType;
import ca.mcscert.jtet.expression.RealVariableType;
import ca.mcscert.jtet.expression.Variable;
import ca.mcscert.jtet.expression.VariableCollection;
import ca.mcscert.jtet.expression.VariableType;

import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.JUnit4;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import static org.hamcrest.CoreMatchers.*;
import static org.junit.Assert.assertThat;

/**
 * @author Matthew Dawson <matthew@mjdsystems.ca>
 */
@RunWith(JUnit4.class)
public class VariableCollectionTest {

    private void AssertsForVariablesAreInBothLists(VariableCollection toTest) {
        assertThat(toTest.getVariables().size(), is(1));
        assertThat(toTest.getVariablesAndEnumeratedValues().size(), is(1));

        assertThat(toTest.getVariables().get("a").name(), is("a"));
        assertThat(toTest.getVariables().get("a").type(), is((VariableType)RealVariableType.Type));
        assertThat(toTest.getVariables(), is(toTest.getVariablesAndEnumeratedValues()));
    }

    @Test
    public void VariablesAreInBothLists() {
        Map<String, Variable> vars = new HashMap<String, Variable>();
        vars.put("a", new Variable("a", new RealVariableType()));

        AssertsForVariablesAreInBothLists(new VariableCollection(vars));
        AssertsForVariablesAreInBothLists(new VariableCollection(vars, null));
        AssertsForVariablesAreInBothLists(new VariableCollection(vars, new HashSet<EnumerationVariableType>()));
    }

    @Test
    public void EnumeratedValuesAreInOneList() {
        EnumerationVariableType type = new EnumerationVariableType("myenum");
        type.enumerationValues().add("e1");
        type.enumerationValues().add("e2");
        type.enumerationValues().add("e3");

        Set<EnumerationVariableType> enums = new HashSet<EnumerationVariableType>();
        enums.add(type);

        VariableCollection vars = new VariableCollection(new HashMap<String, Variable>(), enums);

        assertThat(vars.getVariables().size(), is(0));

        assertThat(vars.getVariablesAndEnumeratedValues().size(), is(3));

        assertThat(vars.getVariablesAndEnumeratedValues().get("e1").name(), is("e1"));
        assertThat(vars.getVariablesAndEnumeratedValues().get("e1").type(), is((VariableType)type));

        assertThat(vars.getVariablesAndEnumeratedValues().get("e2").name(), is("e2"));
        assertThat(vars.getVariablesAndEnumeratedValues().get("e2").type(), is((VariableType)type));

        assertThat(vars.getVariablesAndEnumeratedValues().get("e3").name(), is("e3"));
        assertThat(vars.getVariablesAndEnumeratedValues().get("e3").type(), is((VariableType) type));
    }

    @Test
    public void EnumeratedValuesAndVariablesAreSeparate() {
        EnumerationVariableType type = new EnumerationVariableType("enumtype");
        type.enumerationValues().add("e1");
        type.enumerationValues().add("e2");
        type.enumerationValues().add("e3");

        Set<EnumerationVariableType> enums = new HashSet<EnumerationVariableType>();
        enums.add(type);

        Map<String, Variable> var = new HashMap<String, Variable>();
        var.put("a", new Variable("a", type));

        VariableCollection vars = new VariableCollection(var, enums);

        assertThat(vars.getVariables().size(), is(1));

        assertThat(vars.getVariables().get("a").name(), is("a"));
        assertThat(vars.getVariables().get("a").type(), is((VariableType)type));

        assertThat(vars.getVariablesAndEnumeratedValues().size(), is(4));

        assertThat(vars.getVariablesAndEnumeratedValues().get("a").name(), is("a"));
        assertThat(vars.getVariablesAndEnumeratedValues().get("a").type(), is((VariableType)type));

        assertThat(vars.getVariablesAndEnumeratedValues().get("e1").name(), is("e1"));
        assertThat(vars.getVariablesAndEnumeratedValues().get("e1").type(), is((VariableType)type));

        assertThat(vars.getVariablesAndEnumeratedValues().get("e2").name(), is("e2"));
        assertThat(vars.getVariablesAndEnumeratedValues().get("e2").type(), is((VariableType)type));

        assertThat(vars.getVariablesAndEnumeratedValues().get("e3").name(), is("e3"));
        assertThat(vars.getVariablesAndEnumeratedValues().get("e3").type(), is((VariableType)type));
    }

    @Test(expected = IllegalArgumentException.class)
    public void ConflictingEnumeratedValuesThrow() {
        EnumerationVariableType type = new EnumerationVariableType("myenum");
        type.enumerationValues().add("e1");
        type.enumerationValues().add("e2");
        type.enumerationValues().add("e3");

        Set<EnumerationVariableType> enums = new HashSet<EnumerationVariableType>();
        enums.add(type);

        Map<String, Variable> var = new HashMap<String, Variable>();
        var.put("e1", new Variable("e1", type));

        new VariableCollection(var, enums); // Should throw as e1 the variable conflicts with the enumerated name e1.
    }
}
