/*
 * Copyright (C) 2013 Matthew Dawson <matthew@mjdsystems.ca>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the distribution
 *     * Neither the name of the McMaster Centre for Software Certification nor the names
 *       of its contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
package ca.mcscert.jtet.parsers.test;

import ca.mcscert.jtet.expression.BinaryOperation;
import ca.mcscert.jtet.expression.ExpressionBinaryOperation;
import ca.mcscert.jtet.expression.ExpressionUnaryOperation;
import ca.mcscert.jtet.expression.UnaryOperation;
import ca.mcscert.jtet.expression.VariableCollection;
import ca.mcscert.jtet.parsers.MatlabParser;
import ca.mcscert.jtet.parsers.PVSSimpleParser;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.junit.runners.Parameterized.Parameter;
import org.junit.runners.Parameterized.Parameters;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Arrays;
import java.util.Collection;

import static org.hamcrest.CoreMatchers.is;
import static org.junit.Assert.assertThat;

/** This test simply tests that each operator parses correctly.
 *
 * This is a simply parametrized test set that only deals only with testing the individual operators.  It avoids
 * all the other details (it just uses literals for input).  It also avoids any dependency on generators.
 *
 * @author Matthew Dawson <matthew@mjdsystems.ca>
 */
@RunWith(Parameterized.class)
public class ParserOperatorParseTest {
    @Parameters
    public static Collection<Object[]> data() {
        return Arrays.asList(new Object[][]{
                {"+",  "+",   BinaryOperation.Addition,         null},
                {"-",  "-",   BinaryOperation.Minus,            null},
                {"*",  "*",   BinaryOperation.Multiplication,   null},
                {"/",  "/",   BinaryOperation.Division,         null},
                {"^",  "^",   BinaryOperation.Exponentiation,   null},
                {">",  ">",   BinaryOperation.GreaterThen,      null},
                {">=", ">=",  BinaryOperation.GreaterThenEqual, null},
                {"<",  "<",   BinaryOperation.LessThen,         null},
                {"<=", "<=",  BinaryOperation.LessThenEqual,    null},
                {"==", "=",   BinaryOperation.Equals,           null},
                {"~=", "/=",  BinaryOperation.NotEquals,        null},
                {null, "=>",  BinaryOperation.Implies,          null},
                {"&&", "AND", BinaryOperation.BooleanAnd,       null},
                {"||", "OR",  BinaryOperation.BooleanOr,        null},
                {"~",  "NOT", null,                             UnaryOperation.Negation},
                {"-",  "-",   null,                             UnaryOperation.Minus}
        });
    }
    @Parameter
    public String matlabInputSymbol;

    @Parameter(value = 1)
    public String pvsInputSymbol;

    @Parameter(value = 2)
    public BinaryOperation bop;

    @Parameter(value = 3)
    public UnaryOperation uop;

    private void doTest(String expr, Method method) throws InvocationTargetException, IllegalAccessException {
        if (expr == null) {
            // No symbol for this parser.  Thus skip this test.
            return ;
        }
        if (bop != null) {
            // Binary op to test ...
            ExpressionBinaryOperation exp = (ExpressionBinaryOperation)method.invoke(null, fakeVariableCollection, "1 " + expr + " 2\n");

            assertThat(exp.toString(), is("(1 " + bop.toString() + " 2)"));
        } else {
            // Unary op to test ...
            ExpressionUnaryOperation exp = (ExpressionUnaryOperation)method.invoke(null, fakeVariableCollection, expr + " 3\n");

            assertThat(exp.toString(), is("( " + uop.toString() + " 3)"));
        }
    }

    @Test
    public void testMatlabParser() throws NoSuchMethodException, InvocationTargetException, IllegalAccessException {
        doTest(matlabInputSymbol, MatlabParser.class.getMethod("parseMatlabCode", VariableCollection.class, String.class));
    }

    @Test
    public void testPVSSimpleParser() throws NoSuchMethodException, InvocationTargetException, IllegalAccessException {
        doTest(pvsInputSymbol, PVSSimpleParser.class.getMethod("parsePVSCode", VariableCollection.class, String.class));
    }

    private VariableCollection fakeVariableCollection = new VariableCollection(null);
}
